' Projekteinstellungen:
' Option Strict On
' Option Explicit On
' Option Infer On
' Imports System.Windows.Forms
' Bibliothek "System.Xml.linq" eingebunden 

Imports System.Text
Imports System.Xml
Imports System.Xml.Linq
Imports System.IO

Public Class frmFileDragDrop

    Const _File As String = "FileDragDropData.xml"

#Region "Init"

    Public Sub New()
        InitializeComponent()
        Me.TreeView1.TreeViewNodeSorter = New ComparisonComparer(Of TreeNode)(AddressOf CompareNodes)
    End Sub

    Private Function CompareNodes(ByVal x As TreeNode, ByVal y As TreeNode) As Integer
        'gruppierender Vergleich: zunchst ImageIndex, dann Alphabet
        'dummerweise gibt es 2 Mglichkeiten, den ImageIndex festzulegen: per Key oder per Index
        With Me.ImageList1.Images
            CompareNodes = Math.Max(x.ImageIndex, .IndexOfKey(x.ImageKey)) _
                .CompareTo(Math.Max(y.ImageIndex, .IndexOfKey(y.ImageKey)))
            If CompareNodes <> 0 Then Exit Function 'ImageIndex-Vergleich bringt Ergebnis...
        End With
        Return String.Compare(x.Text, y.Text) '...sonst Stringvergleich returnen
    End Function

#End Region 'Init

#Region "DragDrop-Empfang"

    'Im _DragOver mu durch Setzen von e.Effect festgelegt werden, welcher DropEffect momentan untersttzt wird. Hier: Entweder .None oder .Copy
    'Bei nicht-festsetzen wird DragDropEffects.None angenommen, und es kommt nie zum Dropping
    Private Sub TreeView1_DragOver( _
            ByVal sender As Object, ByVal e As DragEventArgs) Handles TreeView1.DragOver
        With TryCast(e.Data, DataObject)
            If .IsSomething AndAlso .ContainsFileDropList Then
                e.Effect = DragDropEffects.Copy
            End If
        End With
    End Sub

    'Im _DragDrop wird nach dem schluendlich gltigen Drop-Effect verfahren. Dieses Beispiel braucht aber nicht auf verschiedene DropEffekte abzuprfen, da nur DragDropEffects.Copy mglich ist.
    Private Sub TreeView1_DragDrop( _
            ByVal sender As Object, ByVal e As DragEventArgs) _
            Handles TreeView1.DragDrop
        'Ein Problem bei Fehlern whrend des Droppens: Die Standard-Fehlerbehandlung der IDE (Meldung, Codestop) versagt. 
        'Um trotzdem irgendwie debuggen zu knnen hier ein hndischer Codestop, von dem aus im Einzelschritt sich dem Fehler angenhert werden kann
#If DEBUG Then
        For I = 0 To 1
            Try
                ExecDragDrop(e)
                Return
            Catch ex As Exception
                'Fehler aufgetreten. Im Debug-Mode kann der Vorgang im Einzelschrittmodus nochmal durchgangen werden
                Stop
            End Try
        Next
#Else
        Try
            ExecDragDrop(e)
        Catch ex As Exception
            MessageBox.Show( _
                ex.ToString, _
                "Fehler bei der Verarbeitung von TreeView1_DragDrop", _
                MessageBoxButtons.OK, _
                MessageBoxIcon.Error)
        End Try
#End If
    End Sub 'TreeView1_DragDrop


    Private Sub ExecDragDrop(ByVal e As DragEventArgs)

        ' e.Data.GetFileDropList() konvertieren zu: Items As IEnumerable(of String)
        Dim Items = DirectCast(e.Data, DataObject).GetFileDropList.Cast(Of String)()

        ' Pro Itm einen Treenode ermitteln oder erzeugen; den letzten davon sichtbar machen
        Items.Select(Function(Itm) FindOrCreateNode(Itm)).Last.EnsureVisible()

        ''Hm - meist finde ich die Lambda-Schreibweise der Linq-Syntax berlegen:
        'Call (From Itm In Items Select FindOrCreateNode(Itm)).Last.EnsureVisible()
    End Sub


    ''' <summary>
    ''' sucht in TreeView1 nach dem durch FullPath bezeichneten TreeNode.
    ''' bei Nichtfinden einen erstellen, incl. der ihm bergeordneten Nodes
    ''' </summary>
    Private Function FindOrCreateNode(ByVal DroppedFullPath As String) As TreeNode
        Dim Segments = DroppedFullPath.TrimEnd("\"c, " "c).Split("\"c)
        Dim Approach = TreeView1.Nodes.Approach(Segments)
        If Approach.StepsDone = Segments.Length Then Return Approach.Node
        Dim CurrentNodes = If(Approach.Node Is Nothing, TreeView1.Nodes, Approach.Node.Nodes)
        'HandledPath - Diese Nodes existieren schon
        Dim HandledPath = "\".Between(Segments.Take(Approach.StepsDone))

        Dim NewNode As TreeNode = Nothing
        For Each Segment As String In Segments.Skip(Approach.StepsDone) ' diese Nodes noch erzeugen
            HandledPath &= If(HandledPath.Length > 0, "\" & Segment, Segment)
            NewNode = CreateNode(HandledPath)
            'eine Treenode-Ebene tiefer steigen
            CurrentNodes.Add(NewNode)
            CurrentNodes = NewNode.Nodes
        Next
        Return NewNode
    End Function 'EnsureFileSystemItem

#End Region 'DragDrop-Empfang

#Region "DragDrop-'Senden'"

    Private Sub TreeView1_ItemDrag( _
            ByVal sender As Object, ByVal e As ItemDragEventArgs) _
            Handles TreeView1.ItemDrag
        Dim Nd = DirectCast(e.Item, TreeNode)
        Dim DTO As New DataObject
        Dim Lst As New System.Collections.Specialized.StringCollection
        Dim FullPath = Nd.FullPath
        Lst.Add(FullPath)
        DTO.SetFileDropList(Lst)
        TreeView1.AllowDrop = False 'nicht auf sich selber droppen
        ' DoDragDrop ist eine blockierende Function - kehrt also erst zurck, wenn die Maustaste losgelassen wurde
        Dim Executed As DragDropEffects = TreeView1.DoDragDrop(DTO, DragDropEffects.All)
        TreeView1.AllowDrop = True
        Select Case Executed
            Case DragDropEffects.Copy
                'nix tun, kopieren mu die Anwendung, auf die gedropt wurde

            Case DragDropEffects.Move
                'Dieses funktioniert leider nicht.
                'Bei DropEffekt.Move bekomme ich DropEffekt.None zurck
                Nd.Remove()

            Case DragDropEffects.None
                'Also gucken wir, ob das FilesystemItem gelscht wurde
                If (Not File.Exists(FullPath)) AndAlso _
                        (Not Directory.Exists(FullPath)) Then
                    Nd.Remove()
                End If

        End Select
    End Sub

#End Region 'DragDrop-'Senden'

#Region "TreeViewToXml"

    Private Sub BuildTreeNodes(ByVal Xel As XElement, ByVal Nodes As TreeNodeCollection)
        ' aufsteigende Achse umkehren, Root berspringen, Namen sammeln
        Dim Names = Xel.AncestorsAndSelf.Reverse.Skip(1).Select(Function(X) X.@Name)
        ' mit "\" zusammenfgen, daraus einen Treenode erstellen und adden
        Dim Nd = CreateNode("\".Between(Names)).AddTo(Nodes)
        For Each Child In Xel.Elements
            BuildTreeNodes(Child, Nd.Nodes)
        Next
        If Boolean.Parse(Xel.@IsExpanded) Then Nd.Expand()
    End Sub

    Private Sub BuildXml(ByVal Nodes As TreeNodeCollection, ByVal Xel As XElement)
        For Each Nd As TreeNode In Nodes
            'Linq to Xml: ASP - hnliche Syntax
            Dim Child = <Node Name=<%= Nd.Text %> IsExpanded=<%= Nd.IsExpanded %>/>
            Xel.Add(Child)
            BuildXml(Nd.Nodes, Child)
        Next
    End Sub

#End Region 'TreeViewToXml

    ''' <summary>bei RechtsKlick auf einen Node diesen entfernen</summary>
    Private Sub TreeView1_MouseDown( _
            ByVal sender As Object, ByVal e As MouseEventArgs) _
            Handles TreeView1.MouseDown
        If e.Button = Windows.Forms.MouseButtons.Right Then
            With TreeView1.GetNodeAt(e.Location)
                If .IsSomething Then .Remove()
            End With
        End If
    End Sub

    Private Sub MenuItem_Click(ByVal sender As Object, ByVal e As EventArgs) Handles _
            DeleteToolStripMenuItem.Click, _
            ReLoadToolStripMenuItem.Click, _
            SaveToolStripMenuItem.Click
        Select Case True
            Case sender Is DeleteToolStripMenuItem
                File.Delete(_File)

            Case sender Is ReLoadToolStripMenuItem
                Dim FullPath = Path.GetFullPath(_File)
                If Not File.Exists(FullPath) Then
                    MsgBox("'".ConcatWith(FullPath, "'", Lf, Lf, "could not be found"))
                    Return
                End If
                Dim Root = XElement.Load(FullPath)
                Me.TreeView1.Nodes.Clear()
                For Each Child In Root.Elements
                    BuildTreeNodes(Child, Me.TreeView1.Nodes)
                Next

            Case sender Is SaveToolStripMenuItem
                Dim Root = <Root/>
                BuildXml(TreeView1.Nodes, Root)
                Root.Save(_File)

        End Select
    End Sub

    Private Function CreateNode(ByVal FullPath As String) As TreeNode
        Dim NodeName = Path.GetFileName(FullPath)
        Dim sIcon As String
        If NodeName.Length = 0 Then
            NodeName = FullPath
            sIcon = "Drive.Ico"
        ElseIf Directory.Exists(FullPath) Then
            sIcon = "ClsdFold.Ico"
        ElseIf File.Exists(FullPath) Then
            sIcon = "File.Ico"
        Else
            Throw New ArgumentException(FullPath.ConcatWith(Lf, " existiert nicht!"))
        End If
        'VB9: Instanzierung mit With-Klausel ist wesentlich flexibler und leserlicher als die Verwendung (mannigfach) berladener Konstruktoren
        'Dumm nur, da nur Properties angesprochen werden knnen, keine Public Fields
        Return New TreeNode() With { _
            .Text = NodeName, _
            .Name = NodeName, _
            .ImageKey = sIcon, _
            .SelectedImageKey = sIcon}
    End Function

End Class
